#encoding: utf-8
'''
This is a table of TIMIT-and-ARPABET-to-IPA codes.
'''

timitToIPA = {u"ng": u"ŋ",
              u"sh": u"ʃ",
              u"ch": u"tʃ",
              u"y": u"j",
              u"zj": u"ʒ",
              u"zh": u"ʒ",
              u"jh": u"dʒ",
              u"dh": u"ð",
              u"dx": u"ɾ",
              u"tq": u"ʔ",
              u"th": u"θ",
              u"nx": u"n",
              u"p": u"p",
              u"b": u"b",
              u"m": u"m",
              u"t": u"t",
              u"d": u"d",
              u"n": u"n",
              u"k": u"k",
              u"g": u"g",
              u"s": u"s",
              u"z": u"z",
              u"f": u"f",
              u"v": u"v",
              u"w": u"w",
              u"l": u"l",
              u"r": u"r",
              u"aa": u"ɑ",
              u"h": u"h",
              u"hh": u"h",
              u"ae": u"æ",
              u"ah": u"ʌ",
              u"ah0": u"ə",
              u"ah1": u"ʌ",
              u"ao": u"ɔ",
              u"eh": u"ɛ",
              u"ih": u"ɪ",
              u"iy": u"i",
              u"uh": u"ʊ",
              u"uw": u"u",
              u"ay": u"ɑɪ",
              u"aw": u"ɑʊ",
              u"ey": u"eɪ",
              u"ow": u"oʊ",
              u"oy": u"ɔɪ",
              u"er": u"ɚ",
              u"er0": u"ɚ",
              u"er1": u"ɝ",
              u"en": u"n̩",
              u"el": u"l̩",
              # Buckeye contains some vowel + consonant sequences
              u"ihn": u"ɪn",
              u"ayn": u"ɑɪn",
              u"uhn": u"ən",
              u"ehn": u"ɛn",
              u"ehm": u"ɛm",
              u"em": u"ɛm",
              u"aan": u"ɑn",
              }
